"""Provides exceptions relevant to graphs"""


class NodeNotReachableException(Exception):
    """
    Exception for node distance and path
    """

    pass


class GraphModelsCannotBeParsedException(Exception):
    """
    Exception for node distance and path
    """

    pass


class IncorrectUsageException(Exception):
    """
    Exception for node distance and path
    """

    pass

